# -*- coding: utf-8 -*-
"""Re-raise exceptions context manager."""

from typing import TYPE_CHECKING

from six import raise_from

if TYPE_CHECKING:
    from typing import Optional, Tuple, Type, Union

__all__ = ["ReraiseContext"]


class ReraiseContext(object):
    """
    Re-raise exceptions context manager for shorter tracebacks and custom messages.

    .. code:: python

        >>> from objetto.utils.reraise_context import ReraiseContext

        >>> def func_a():
        ...     func_b()
        ...
        >>> def func_b():
        ...     func_c()
        ...
        >>> def func_c():
        ...     raise ValueError("something is wrong")
        ...
        >>> with ReraiseContext(ValueError, "oops"):
        ...     func_a()
        Traceback (most recent call last):
        ValueError: oops; something is wrong

    :param exc_type: Exception type(s) to catch.
    :type exc_type: type[Exception] or tuple[type[Exception]]

    :param message: Optional additional custom message.
    :type message: str
    """

    __slots__ = ("__exc_type", "__message")

    def __init__(
        self,
        exc_type=Exception,  # type: Union[Type[Exception], Tuple[Type[Exception], ...]]
        message=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self.__exc_type = exc_type
        self.__message = message

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            if not isinstance(exc_val, self.exc_type):
                return False
            exc_message = "{}".format(exc_val)
            if self.message is not None:
                if exc_message:
                    exc_message = "{}; {}".format(self.message, exc_message)
                else:
                    exc_message = self.message
            exc = type(exc_val)(exc_message, *exc_val.args[1:])
            raise_from(exc, None)
        return False

    @property
    def exc_type(self):
        # type: () -> Union[Type[Exception], Tuple[Type[Exception], ...]]
        """
        Exception type(s) to catch.

        :rtype: type[Exception] or tuple[type[Exception]]
        """
        return self.__exc_type

    @property
    def message(self):
        # type: () -> Optional[str]
        """
        Optional additional custom message.

        :rtype: str or None
        """
        return self.__message



# -*- coding: utf-8 -*-
"""Recursive-ready `repr` decorator."""

from collections import Counter as ValueCounter
from threading import local
from typing import TYPE_CHECKING

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc  # type: ignore

from decorator import decorator
from qualname import qualname  # type: ignore

if TYPE_CHECKING:
    from typing import Any, Callable, Optional

__all__ = ["recursive_repr"]

_local = local()


@decorator
def recursive_repr(
    func,  # type: Callable
    max_depth=1,  # type: Optional[int]
    max_global_depth=2,  # type: Optional[int]
    *args,  # type: Any
    **kwargs  # type: Any
):
    # type: (...) -> str
    """
    Decorate a representation method/function and prevents infinite recursion.

    .. code:: python

        >>> from objetto.utils.recursive_repr import recursive_repr

        >>> class MyClass(object):
        ...
        ...     @recursive_repr
        ...     def __repr__(self):
        ...         return "MyClass<" + repr(self) + ">"
        ...
        >>> my_obj = MyClass()
        >>> repr(my_obj)
        'MyClass<...>'

    :param func: The '__repr__' and/or '__str__' method/function.
    :type func: function

    :param max_depth: Maximum recursion depth before returning fill value.
    :type max_depth: int

    :param max_global_depth: Maximum global depth before returning fill value.
    :type max_global_depth: int

    :return: Decorated method function.
    :rtype: function
    """
    self = args[0]
    self_id = id(self)
    try:
        reprs = _local.reprs
    except AttributeError:
        reprs = _local.reprs = ValueCounter()
    try:
        global_reprs = _local.global_reprs
    except AttributeError:
        global_reprs = _local.global_reprs = [0]

    reprs[self_id] += 1
    global_reprs[0] += 1
    try:
        if max_depth is not None and reprs[self_id] > max_depth:
            return "..."
        elif max_global_depth is not None and global_reprs[0] > max_global_depth:
            try:
                qual_name = qualname(type(self))
            except AttributeError:
                qual_name = type(self).__name__
            return "<{} at {}>".format(qual_name, hex(id(self)))
        else:
            return func(*args, **kwargs)
    finally:
        reprs[self_id] -= 1
        if not reprs[self_id]:
            del reprs[self_id]
        if not reprs:
            del _local.reprs
        global_reprs[0] -= 1
        if not global_reprs[0]:
            del _local.global_reprs



# -*- coding: utf-8 -*-
"""Base classes and metaclasses."""

from .bases import (
    ABSTRACT_TAG,
    FINAL_CLASS_TAG,
    FINAL_METHOD_TAG,
    INITIALIZING_TAG,
    MISSING,
    Base,
    BaseCollection,
    BaseContainer,
    BaseHashable,
    BaseInteractiveCollection,
    BaseIterable,
    BaseMeta,
    BaseMutableCollection,
    BaseProtectedCollection,
    BaseSized,
    Generic,
    abstract_member,
    final,
    init,
    init_context,
    make_base_cls,
    simplify_member_names,
)
from .dict import BaseDict, BaseInteractiveDict, BaseMutableDict, BaseProtectedDict
from .list import BaseInteractiveList, BaseList, BaseMutableList, BaseProtectedList
from .set import BaseInteractiveSet, BaseMutableSet, BaseProtectedSet, BaseSet

__all__ = [
    "MISSING",
    "ABSTRACT_TAG",
    "FINAL_CLASS_TAG",
    "FINAL_METHOD_TAG",
    "INITIALIZING_TAG",
    "final",
    "init_context",
    "init",
    "simplify_member_names",
    "make_base_cls",
    "BaseMeta",
    "Base",
    "abstract_member",
    "Generic",
    "BaseHashable",
    "BaseSized",
    "BaseIterable",
    "BaseContainer",
    "BaseCollection",
    "BaseProtectedCollection",
    "BaseInteractiveCollection",
    "BaseMutableCollection",
    "BaseDict",
    "BaseProtectedDict",
    "BaseInteractiveDict",
    "BaseMutableDict",
    "BaseList",
    "BaseProtectedList",
    "BaseInteractiveList",
    "BaseMutableList",
    "BaseSet",
    "BaseProtectedSet",
    "BaseInteractiveSet",
    "BaseMutableSet",
]



# -*- coding: utf-8 -*-
"""Observer mixin class."""

from abc import abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type, cast
from weakref import ref

from ._applications import Action, Phase
from ._bases import final
from ._objects import BaseObject
from .data import Data, data_attribute
from .utils.reraise_context import ReraiseContext
from .utils.subject_observer import Observer, ObserverToken
from .utils.type_checking import assert_is_instance

if TYPE_CHECKING:
    from typing import Any

    from .data import DataAttribute

__all__ = ["ActionObserver", "ActionObserverToken", "ActionObserverExceptionData"]


class InternalObserver(Observer):
    """
    Internal observer.

    :param action_observer: Action observer.
    """

    def __init__(self, action_observer):
        # type: (ActionObserver) -> None
        self.action_observer_ref = ref(action_observer)

    def __observe__(self, *payload):
        # type: (Any) -> None
        """
        Receive payload, unpack it, and relay it to the action observer.

        :param payload: Payload.
        """
        action_observer = self.action_observer_ref()
        if action_observer is not None:
            action, phase = payload
            action_observer.__observe__(action, phase)


# noinspection PyAbstractClass
class ActionObserver(object):
    """
    Mixin/abstract class for observing an object for actions.

    .. code:: python

        >>> from objetto import Application, Object, attribute
        >>> from objetto.observers import ActionObserver

        >>> class Person(Object):
        ...     name = attribute(str, default="Albert")
        ...
        >>> class PersonObserver(ActionObserver):
        ...
        ...     def __observe__(self, action, phase):
        ...         print((action.change.name, phase.value))
        ...
        >>> app = Application()
        >>> person = Person(app)
        >>> observer = PersonObserver()
        >>> token = observer.start_observing(person)
        >>> person.name = "Einstein"
        ('Update Attributes', 'PRE')
        ('Update Attributes', 'POST')
    """

    __internal_observer = None  # type: Optional[InternalObserver]

    @abstractmethod
    def __observe__(self, action, phase):
        # type: (Action, Phase) -> None
        """
        Observe an action (and its execution phase) from an object.

        :param action: Action.
        :type action: objetto.objects.Action

        :param phase: Phase.
        :type phase: :data:`objetto.constants.PRE` or :data:`objetto.constants.POST`

        :raises NotImplementedError: Abstract method not implemented.
        """
        error = (
            "object action observer class '{}' did not implement abstract method "
            "'__observe__'; can't observe action {} during {}"
        ).format(type(self).__name__, action, phase)
        raise NotImplementedError(error)

    def start_observing(self, obj):
        # type: (BaseObject) -> ActionObserverToken
        """
        Start observing an object for actions.

        :param obj: Object.
        :type obj: objetto.objects.Object

        :return: Observer token.
        :rtype: objetto.observers.ActionObserverToken

        :raises TypeError: Invalid 'obj' parameter type.
        :raises RuntimeError: Can't start observing while object is initializing.
        """
        with ReraiseContext(TypeError, "'obj' parameter"):
            assert_is_instance(obj, BaseObject)
        if obj._initializing:
            error = "can't start observing object {} during its initialization"
            raise RuntimeError(error)
        obj.__.subject.register_observer(self.__observer)
        action_observer_token = cast(
            "ActionObserverToken",
            ActionObserverToken.__make__(obj.__.subject, self.__observer),
        )
        return action_observer_token

    def stop_observing(self, obj):
        # type: (BaseObject) -> None
        """
        Stop observing an object for actions.

        :param obj: Object.
        :type obj: objetto.objects.Object

        :raises TypeError: Invalid 'obj' parameter type.
        """
        with ReraiseContext(TypeError, "'obj' parameter"):
            assert_is_instance(obj, BaseObject)
        obj.__.subject.deregister_observer(self.__observer)

    @property
    def __observer(self):
        # type: () -> InternalObserver
        """Internal observer."""
        if self.__internal_observer is None:
            self.__internal_observer = InternalObserver(self)
        return self.__internal_observer


# noinspection PyAbstractClass
class ActionObserverToken(ObserverToken):
    """
    Allows control over observers' order/priority.

    Inherits from:
      - :class:`objetto.utils.subject_observer.ObserverToken`

    .. note::
        This class can't be instantiated directly. A token should be retrieved when an
        action observer starts observing an object.

    .. code:: python

        >>> from objetto import Application, Object, attribute
        >>> from objetto.constants import POST
        >>> from objetto.observers import ActionObserver

        >>> observed_names = []
        ...
        >>> class Person(Object):
        ...     name = attribute(str, default="Albert")
        ...
        >>> class PersonObserver(ActionObserver):
        ...     def __init__(self, this, dependency=None):
        ...         self.this = this
        ...         self.dependency = dependency
        ...
        ...     def __observe__(self, action, phase):
        ...         if self.dependency:
        ...             self.dependency.wait()
        ...         if phase is POST:
        ...             observed_names.append(
        ...                 "{} {}".format(self.this, action.change.new_values["name"])
        ...             )
        ...
        >>> app = Application()
        >>> person = Person(app)

        >>> observer_a = PersonObserver("A")
        >>> token_a = observer_a.start_observing(person)

        >>> observer_b = PersonObserver("B", dependency=token_a)
        >>> token_b = observer_b.start_observing(person)

        >>> person.name = "Einstein"
        >>> observed_names  # note how A will always observe the change before B does
        ['A Einstein', 'B Einstein']
    """

    __slots__ = ()

    @property
    def observer(self):
        # type: () -> Optional[ActionObserver]
        """
        Action observer.

        :rtype: objetto.observers.ActionObserver or None
        """
        internal_observer = self._observer_ref()
        if internal_observer is None:
            return None
        else:
            return cast("InternalObserver", internal_observer).action_observer_ref()


@final
class ActionObserverExceptionData(Data):
    """
    Describes an exception raised by an observer while observing an action.

    Inherits from:
      - :class:`objetto.data.Data`
    """

    observer = data_attribute(
        ActionObserver, checked=False
    )  # type: DataAttribute[ActionObserver]
    """
    Action observer.

    :type: objetto.observers.ActionObserver
    """

    action = data_attribute(Action, checked=False)  # type: DataAttribute[Action]
    """
    Action.

    :type: objetto.objects.Action
    """

    phase = data_attribute(Phase, checked=False)  # type: DataAttribute[Phase]
    """
    Phase.

    :type: :data:`objetto.constants.PRE` or :data:`objetto.constants.POST`
    """

    exception_type = data_attribute(
        (type(BaseException), None), checked=False
    )  # type: DataAttribute[Optional[Type[BaseException]]]
    """
    Exception type.

    :type: type[BaseException] or None
    """

    exception = data_attribute(
        (BaseException, None), checked=False
    )  # type: DataAttribute[Optional[BaseException]]
    """
    Exception.

    :type: BaseException or None
    """

    traceback = data_attribute(
        (TracebackType, None), checked=False
    )  # type: DataAttribute[Optional[TracebackType]]
    """
    Traceback.

    :type: types.TracebackType or None
    """



# -*- coding: utf-8 -*-
"""Manages multiple objects under different contexts."""

from collections import Counter as ValueCounter
from contextlib import contextmanager
from copy import deepcopy
from enum import Enum, unique
from inspect import getmro
from threading import RLock
from traceback import format_exception
from typing import TYPE_CHECKING, TypeVar, cast, overload
from weakref import WeakKeyDictionary

from six import iteritems, itervalues, string_types, with_metaclass

from ._bases import Base, BaseMeta, Generic, final
from ._changes import BaseChange
from ._data import BaseData, DataAttribute, InteractiveDictData
from ._exceptions import BaseObjettoException
from ._states import BaseState, DictState
from .data import (
    Data,
    DictData,
    InteractiveData,
    ListData,
    data_attribute,
    data_dict_attribute,
    data_list_attribute,
    data_set_attribute,
)
from .utils.custom_repr import custom_mapping_repr
from .utils.recursive_repr import recursive_repr
from .utils.reraise_context import ReraiseContext
from .utils.type_checking import (
    assert_is_callable,
    assert_is_instance,
    assert_is_subclass,
)
from .utils.weak_reference import WeakReference

if TYPE_CHECKING:
    from typing import (
        AbstractSet,
        Any,
        Callable,
        Counter,
        Dict,
        Final,
        Iterator,
        List,
        Mapping,
        MutableMapping,
        Optional,
        Set,
        Tuple,
        Type,
        Union,
    )

    from ._changes import BaseAtomicChange, Batch
    from ._data import InteractiveSetData
    from ._history import HistoryObject
    from ._objects import BaseObject, Relationship
    from ._observers import ActionObserverExceptionData, InternalObserver
    from .utils.subject_observer import ObserverExceptionInfo

    assert Relationship
    assert InternalObserver

    ReadFunction = Callable[[], "Store"]
    WriteFunction = Callable[
        [Any, BaseData, Mapping[str, Any], Counter[BaseObject], BaseAtomicChange], None
    ]

    ReadMetadataFunction = Callable[[], InteractiveDictData]
    UpdateMetadataFunction = Callable[[Mapping[str, Any]], None]

__all__ = [
    "ActionObserversFailedError",
    "RejectChangeException",
    "Phase",
    "Action",
    "Store",
    "BO",
    "ApplicationRoot",
    "Application",
]


@unique
class Phase(Enum):
    """Action phase."""

    PRE = "PRE"
    """Before the changes."""

    POST = "POST"
    """After the changes."""


class ActionObserversFailedError(BaseObjettoException):
    """
    Action observers failed while observing action.

    Inherits from:
      - :class:`objetto.bases.BaseObjettoException`

    :param exception_infos: Observer exception infos.
    :type exception_infos: tuple[objetto.observers.ActionObserverExceptionData]
    """

    def __init__(self, message, exception_infos):
        # type: (str, Tuple[ActionObserverExceptionData, ...]) -> None
        message = (
            (
                message
                + "\n\n"
                + "\n".join(
                    (
                        ("Observer: {}\n" "Change: {}\n" "Phase: {}\n").format(
                            exception_info.observer,
                            type(exception_info.action.change).__fullname__,
                            exception_info.phase.name,
                        )
                        + "".join(
                            format_exception(
                                exception_info.exception_type,
                                exception_info.exception,
                                exception_info.traceback,
                            )
                        )
                    )
                    for exception_info in exception_infos
                )
            )
            if exception_infos
            else message
        )
        super(ActionObserversFailedError, self).__init__(message)
        self.__exception_infos = exception_infos

    @property
    def exception_infos(self):
        # type: () -> Tuple[ActionObserverExceptionData, ...]
        """
        Observer exception infos.

        :rtype: tuple[objetto.observers.ActionObserverExceptionData]
        """
        return self.__exception_infos


class RejectChangeException(BaseObjettoException):
    """
    Exception to be raised from within a reaction. This will cause the change to be
    reverted and and the custom callback function to run after that.

    Inherits from:
      - :class:`objetto.bases.BaseObjettoException`

    :param change: Change to reject.
    :type change: objetto.bases.BaseChange

    :param callback: Callback to run after change is rewound.
    :type callback: function or collections.abc.Callable
    """

    def __init__(self, message, change, callback):
        # type: (str, BaseChange, Callable[[], None]) -> None
        from ._changes import BaseChange

        with ReraiseContext(TypeError, "'change' parameter"):
            assert_is_instance(change, BaseChange)

        with ReraiseContext(TypeError, "'callback' parameter"):
            assert_is_callable(callback)

        message = (
            "{}; change {} was rejected but callback could not run because rejection "
            "was not raised and/or caught within the correct context"
        ).format(message, change)
        super(RejectChangeException, self).__init__(message)
        self.__change = change
        self.__callback = callback

    @property
    def change(self):
        # type: () -> BaseChange
        """
        Change to reject.

        :rtype: objetto.bases.BaseChange
        """
        return self.__change

    @property
    def callback(self):
        # type: () -> Callable[[], None]
        """
        Callback to run after change is rewound.

        :rtype: function or collections.abc.Callable
        """
        return self.__callback


class TemporaryContextException(BaseObjettoException):
    """Temporary write context exception."""


class ApplicationLock(Base):
    """
    Re-entrant threading lock for thread-safe applications.

      - Can be deep copied and pickled.
    """

    __slots__ = ("__lock",)

    def __init__(self):
        # type: () -> None
        self.__lock = RLock()

    def __deepcopy__(self, memo=None):
        # type: (Optional[Dict[int, Any]]) -> ApplicationLock
        """
        Make a deep copy.

        :param memo: Memo dict.
        :return: Deep copy.
        """
        if memo is None:
            memo = {}
        try:
            deep_copy = memo[id(self)]
        except KeyError:
            deep_copy = memo[id(self)] = type(self)()
        return deep_copy

    def __enter__(self):
        """Enter lock context."""
        return self.__lock.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit lock context."""
        return self.__lock.__exit__(exc_type, exc_val, exc_tb)

    def __reduce__(self):
        # type: () -> Tuple[Type[ApplicationLock], Tuple]
        """
        Reduce for pickling.

        :return: Class and init arguments.
        """
        return type(self), ()


class ApplicationStorage(WeakKeyDictionary):
    """
    Application storage.

      - Holds a store for each object in the application.
      - Can be deep copied and pickled.
    """

    def __deepcopy__(self, memo=None):
        # type: (Optional[Dict[int, Any]]) -> ApplicationStorage
        """
        Make a deep copy.

        :param memo: Memo dict.
        :return: Deep copy.
        """
        if memo is None:
            memo = {}
        try:
            deep_copy = memo[id(self)]
        except KeyError:
            deep_copy = memo[id(self)] = type(self)()
            deep_copy_args = dict(self), memo
            deep_copy.update(deepcopy(*deep_copy_args))
        return deep_copy

    def __reduce__(self):
        # type: () -> Tuple[Type[ApplicationStorage], Tuple[Dict[BaseObject, Store]]]
        """
        Reduce for pickling.

        :return: Class and init arguments.
        """
        return type(self), (dict(self),)


class Store(InteractiveData):
    """Holds an object's state, data, metadata, hierarchy, and history information."""

    state = data_attribute(
        BaseState, subtypes=True, checked=False
    )  # type: DataAttribute[BaseState]
    """State."""

    data = data_attribute(
        (BaseData, None), subtypes=True, checked=False
    )  # type: DataAttribute[Optional[BaseData]]
    """Data."""

    metadata = cast(
        "DataAttribute[InteractiveDictData[str, Any]]",
        data_dict_attribute(key_types=string_types, checked=False),
    )  # type: DataAttribute[InteractiveDictData[str, Any]]
    """Metadata."""

    parent_ref = data_attribute(
        cast("Type[WeakReference[BaseObject]]", WeakReference),
        checked=False,
        default=WeakReference(),
    )  # type: DataAttribute[WeakReference[BaseObject]]
    """Weak reference to the parent."""

    history_provider_ref = data_attribute(
        cast("Type[WeakReference[BaseObject]]", WeakReference),
        checked=False,
        default=WeakReference(),
    )  # type: DataAttribute[WeakReference[BaseObject]]
    """Weak reference to the history provider."""

    last_parent_history_ref = data_attribute(
        cast("Type[WeakReference[HistoryObject]]", WeakReference),
        checked=False,
        default=WeakReference(),
    )  # type: DataAttribute[WeakReference[HistoryObject]]
    """Weak reference to the last history object."""

    history = data_attribute(
        (".._history|HistoryObject", None), checked=False, default=None
    )  # type: DataAttribute[HistoryObject]
    """History object."""

    children = cast(
        "DataAttribute[InteractiveSetData[BaseObject]]",
        data_set_attribute(".._objects|BaseObject", subtypes=True, checked=False),
    )  # type: DataAttribute[InteractiveSetData[BaseObject]]
    """Children."""


@final
class Action(Data):
    """
    Carries information about a change and where it happened in the hierarchy.

    Inherits from:
      - :class:`objetto.data.Data`
    """

    sender = data_attribute(
        ".._objects|BaseObject", subtypes=True, checked=False
    )  # type: DataAttribute[BaseObject]
    """
    Object where the action originated from (where the change happened).

    :type: objetto.bases.BaseObject
    """

    receiver = data_attribute(
        ".._objects|BaseObject", subtypes=True, checked=False
    )  # type: DataAttribute[BaseObject]
    """
    Object relaying the action up the hierarchy.

    :type: objetto.bases.BaseObject
    """

    locations = cast(
        "DataAttribute[ListData[Any]]",
        data_list_attribute(checked=False, interactive=False),
    )  # type: DataAttribute[ListData[Any]]
    """
    List of relative locations from the receiver to the sender.

    :type: list[str or int or collections.abc.Hashable]
    """

    change = data_attribute(
        BaseChange, subtypes=True, checked=False
    )  # type: DataAttribute[BaseChange]
    """
    Change that happened in the sender.

    :type: objetto.bases.BaseChange
    """


class Commit(Data):
    """Holds unmerged, modified stores."""

    actions = cast(
        "DataAttribute[ListData[Action]]",
        data_list_attribute(Action, checked=False, finalized=True, interactive=False),
    )  # type: Final[DataAttribute[ListData[Action]]]
    """Actions."""

    stores = cast(
        "DataAttribute[DictData[BaseObject, Store]]",
        data_dict_attribute(
            Store,
            checked=False,
            key_types=".._objects|BaseObject",
            key_subtypes=True,
            interactive=False,
        ),
    )  # type: Final[DataAttribute[DictData[BaseObject, Store]]]
    """Modified stores."""


@final
class BatchCommit(Commit):
    """Batch commit."""

    phase = data_attribute(Phase, checked=False)  # type: DataAttribute[Phase]
    """Batch phase."""


# noinspection PyTypeChecker
_AR = TypeVar("_AR", bound="ApplicationRoot")

# noinspection PyTypeChecker
BO = TypeVar("BO", bound="BaseObject")


@final
class ApplicationRoot(Base, Generic[BO]):
    """
    Describes a root object that gets initialized with the application.

    .. note::
        Prefer using the :func:`objetto.applications.root` factory over
        instantiating :class:`objetto.applications.ApplicationRoot` directly.

    :param obj_type: Object type.
    :type obj_type: type[objetto.bases.BaseObject]

    :param priority: Initialization priority.
    :type priority: int or None

    :param kwargs: Keyword arguments to be passed to the object's `__init__`.

    :raises ValueError: Used reserved keyword argument.
    :raises TypeError: Invalid object type.
    """

    __slots__ = ("__obj_type", "__priority", "__kwargs")

    def __init__(self, obj_type, priority=None, **kwargs):
        # type: (Type[BO], Optional[int], Any) -> None

        # Check kwargs for reserved keys.
        if "app" in kwargs:
            error = "can't use reserved keyword argument 'app'"
            raise ValueError(error)

        from ._objects import BaseObject

        with ReraiseContext(TypeError, "'obj_type' parameter"):
            assert_is_subclass(obj_type, BaseObject)
        self.__obj_type = obj_type
        self.__priority = priority
        self.__kwargs = DictState(kwargs)

    @overload
    def __get__(self, instance, owner):
        # type: (_AR, None, Type[Application]) -> _AR
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (Application, Type[Application]) -> BO
        pass

    @overload
    def __get__(self, instance, owner):
        # type: (_AR, object, type) -> _AR
        pass

    def __get__(self, instance, owner):
        """
        Get attribute value when accessing from valid instance or this descriptor
        otherwise.

        :param instance: Instance.
        :type instance: objetto.applications.Application or None

        :param owner: Owner class.
        :type owner: type[objetto.applications.Application]

        :return: Object instance or this descriptor.
        :rtype: objetto.bases.BaseObject or objetto.applications.ApplicationRoot
        """
        if instance is not None and isinstance(instance, Application):
            return instance.__.get_root_obj(self)
        return self

    def __hash__(self):
        # type: () -> int
        """
        Get hash based on object id.

        :return: Hash based on object id.
        :rtype: int
        """
        return hash(id(self))

    def __eq__(self, other):
        # type: (object) -> bool
        """
        Compare with another object for identity.

        :param other: Another object.

        :return: True if the same object.
        :rtype: bool
        """
        return other is self

    @recursive_repr
    def __repr__(self):
        # type: () -> str
        """
        Get representation.

        :return: Representation.
        :rtype: str
        """
        return custom_mapping_repr(
            self.to_dict(),
            prefix="{}(".format(type(self).__name__),
            template="{key}={value}",
            suffix=")",
            key_repr=str,
        )

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """
        Convert to dictionary.

        :return: Dictionary.
        :rtype: dict[str, Any]
        """
        return {
            "obj_type": self.obj_type,
            "priority": self.priority,
            "kwargs": self.kwargs,
        }

    @property
    def obj_type(self):
        # type: () -> Type[BO]
        """
        Object type.

        :rtype: type[objetto.bases.BaseObject]
        """
        return self.__obj_type

    @property
    def priority(self):
        # type: () -> Optional[int]
        """
        Initialization priority.

        :rtype: int or None
        """
        return self.__priority

    @property
    def kwargs(self):
        # type: () -> DictState[str, Any]
        """Keyword arguments to be passed to object's '__init__'."""
        return self.__kwargs


class ApplicationInternals(Base):
    """Internals for `Application`."""

    __slots__ = (
        "__app_ref",
        "__history_cls",
        "__lock",
        "__storage",
        "__busy_writing",
        "__busy_hierarchy",
        "__commits",
        "__reading",
        "__writing",
        "__roots",
    )

    def __init__(self, app):
        # type: (Application) -> None
        self.__app_ref = WeakReference(app)
        self.__history_cls = None  # type: Optional[Type[HistoryObject]]
        self.__lock = ApplicationLock()
        self.__storage = ApplicationStorage()
        self.__busy_writing = set()  # type: Set[BaseObject]
        self.__busy_hierarchy = ValueCounter()  # type: Counter[BaseObject]
        self.__commits = []  # type: List[Commit]
        self.__reading = []  # type: List[Optional[BaseObject]]
        self.__writing = []  # type: List[Optional[BaseObject]]
        self.__roots = {}  # type: Dict[ApplicationRoot, BaseObject]

    def __deepcopy__(self, memo=None):
        # type: (Optional[Dict[int, Any]]) -> ApplicationInternals
        """
        Deep copy.

        :param memo: Memo dict.
        :return: Deep copy.
        :raises RuntimeError: Can't deep copy while application is in a 'write' context.
        """
        if memo is None:
            memo = {}
        with self.read_context():
            if self.__writing:
                error = "can't deep copy while application is in a 'write' context"
                raise RuntimeError(error)
            try:
                deep_copy = memo[id(self)]
            except KeyError:
                cls = type(self)
                deep_copy = memo[id(self)] = cls.__new__(cls)
                deep_copy_state_args = self.__getstate__(), memo
                deep_copy_state = deepcopy(*deep_copy_state_args)
                deep_copy.__setstate__(deep_copy_state)
            return deep_copy

    def __getstate__(self):
        # type: () -> Dict[str, Any]
        """
        Get state for pickling.

        :return: State.
        :raises RuntimeError: Can't pickle while application is in a 'write' context.
        """
        with self.read_context():
            if self.__writing:
                error = "can't pickle while application is in a 'write' context"
                raise RuntimeError(error)
            return super(ApplicationInternals, self).__getstate__()

    def __read(self, obj):
        # type: (BaseObject) -> Store
        """
        Get current store for an object.

        :param obj: Object.
        :return: Store.
        """
        if self.__writing:
            try:
                return self.__commits[-1].stores[obj]
            except (IndexError, KeyError):
                pass
        try:
            return self.__storage[obj]
        except KeyError:
            error = "object {} is no longer valid".format(obj)
            raise RuntimeError(error)

    def __read_history(
        self,
        obj,  # type: BaseObject
    ):
        # type: (...) -> Union[Tuple[HistoryObject, BaseObject], Tuple[None, None]]
        """
        Get current history for an object.

        :param obj: Object.
        :return: History object and history provider or (None, None).
        """
        store = self.__read(obj)
        if store.history is not None:
            return store.history, obj
        provider = store.history_provider_ref()
        if provider is not None:
            return self.__read_history(provider)
        return None, None

    def __pre_parent_check(
        self,
        obj,  # type: BaseObject
        hierarchy,  # type: List[BaseObject]
        child_counter,  # type: Counter[BaseObject]
    ):
        # type: (...) -> None
        """
        Run checks before performing parenting/unparenting.

        :param obj: Object adopting/releasing children.
        :param hierarchy: Cached hierarchy.
        :param child_counter: Child object counter.
        :raises ValueError: Can't have history objects as children of other objects.
        :raises ValueError: Can't have root objects as children of other objects.
        :raises ValueError: Can't change parent while object's hierarchy is locked.
        :raises ValueError: Can't change parent while object is initializing.
        :raises ValueError: Object is already parented.
        :raises ValueError: Parent cycle detected.
        :raises ValueError: Object is not a child.
        :raises ValueError: Can't parent more than once.
        :raises ValueError: Can't unparent more than once.
        """
        if not child_counter:
            return
        children = None
        for child, count in iteritems(child_counter):
            if count:
                if self.__history_cls is not None:
                    if isinstance(child, self.__history_cls):
                        error = (
                            "can't have '{}' objects as children of other objects"
                        ).format(self.__history_cls.__fullname__)
                        raise ValueError(error)
                if child.__.is_root:
                    error = "'{}' object is a root and can't be parented".format(
                        type(child).__fullname__
                    )
                    raise ValueError(error)
                if self.__busy_hierarchy.get(child):
                    error = (
                        "can't change parent for {} while its hierarchy is locked"
                    ).format(child)
                    raise ValueError(error)
                if child._initializing:
                    error = (
                        "can't change parent for {} while running it's '__init__'"
                    ).format(child)
                    raise ValueError(error)
            if count == 1:
                child_parent = self.__read(child).parent_ref()
                if child_parent is not None:
                    error = (
                        "{} is already parented to {}, can't parent it to {}"
                    ).format(child, child_parent, obj)
                    raise ValueError(error)
                for parent in hierarchy:
                    if parent is child:
                        error = "parent cycle between {} and {}".format(child, obj)
                        raise ValueError(error)
            elif count == -1:
                if children is None:
                    children = self.__read(obj).children
                if child not in children:
                    error = "{} is not a child of {}".format(child, obj)
                    raise ValueError(error)
            elif count > 1:
                error = "{} can't be parented to {} more than once".format(child, obj)
                raise ValueError(error)
            elif count < -1:
                error = "{} can't be un-parented from {} more than once".format(
                    child, obj
                )
                raise ValueError(error)

    def __write(
        self,
        obj,  # type: BaseObject
        state,  # type: BaseState
        data,  # type: Optional[BaseData]
        metadata,  # type: Mapping[str, Any]
        child_counter,  # type: Counter[BaseObject]
        change,  # type: BaseAtomicChange
    ):
        # type: (...) -> None
        """
        Perform a 'write' operation.

        :param obj: Object.
        :param state: New state.
        :param data: New data.
        :param metadata: New metadata.
        :param child_counter: Child counter.
        :param change: Change.
        """

        # Lock hierarchy and cache it.
        with self.__hierarchy_context(obj) as hierarchy:
            assert hierarchy[0] is obj

            # Perform pre-parent check.
            self.__pre_parent_check(obj, hierarchy, child_counter)

            # Lock parenting for new children.
            with self.__new_children_context(change.new_children):

                # Enter history atomic batch if not in a batch.
                history, history_provider = self.__read_history(obj)
                atomic_batch_change = None
                if (
                    history is not None
                    and not obj._initializing
                    and not history.executing
                    and not history.in_batch()
                ):
                    from ._changes import Batch

                    atomic_batch_change = Batch(
                        name=change.name,
                        obj=obj,
                        metadata={"atomic_change": change, "is_atomic": True},
                    )
                    history.__enter_batch__(atomic_batch_change)

                # Pre phase.
                child = None  # type: Optional[BaseObject]
                single_locations = []  # type: List[Any]
                single_data_locations = []  # type: List[Any]
                all_locations = []  # type: List[List[Any]]
                all_data_locations = []  # type: List[List[Any]]
                actions = []  # type: List[Action]
                for i, parent in enumerate(hierarchy):
                    if i == 0:
                        location = None
                        locations = []
                        data_location = None
                        data_locations = []
                    else:
                        assert child is not None
                        location = parent._locate(child)
                        locations = [location] + all_locations[-1]

                        relationship = cast(
                            "Relationship", parent._get_relationship(location)
                        )
                        if relationship.data:
                            data_location = parent._locate_data(child)
                            data_locations = [data_location, all_data_locations[-1]]
                        else:
                            data_location = None
                            data_locations = []

                    single_locations.append(location)
                    all_locations.append(locations)

                    single_data_locations.append(data_location)
                    all_data_locations.append(data_locations)

                    action = Action(
                        sender=obj,
                        receiver=parent,
                        locations=locations,
                        change=change,
                    )
                    assert action.sender is obj
                    assert action.receiver is parent

                    actions.append(action)
                    self.__react(action.receiver, action, Phase.PRE)

                    child = parent

                # Flush histories and filter history adopters.
                new_children_last_parent_history_updates = set()
                histories_to_flush = set()
                filtered_history_adopters = set()

                # noinspection PyTypeChecker
                for new_child in change.new_children:
                    new_child_store = self.__read(new_child)
                    last_parent_history = new_child_store.last_parent_history_ref()
                    if last_parent_history is not history:
                        if last_parent_history is not None:
                            histories_to_flush.add(last_parent_history)
                        new_children_last_parent_history_updates.add(new_child)

                # noinspection PyTypeChecker
                for adopter in change.history_adopters:
                    if type(adopter)._history_descriptor is not None:
                        continue
                    filtered_history_adopters.add(adopter)
                    adopter_old_history, _ = self.__read_history(adopter)
                    if adopter_old_history is not history:
                        if adopter_old_history is not None:
                            histories_to_flush.add(adopter_old_history)

                for history_to_flush in histories_to_flush:
                    history_to_flush.flush()

                # Store changes.
                try:
                    stores = self.__commits[-1].stores
                except IndexError:
                    stores = InteractiveDictData()

                store = self.__read(obj)
                old_data = store.data
                store = store.update(
                    {
                        "state": state,
                        "data": data,
                        "metadata": metadata,
                    }
                )

                # Children changes.
                if change.old_children or change.new_children:
                    children = store.children

                    # noinspection PyTypeChecker
                    for old_child in change.old_children:
                        children = children.remove(old_child)
                        child_store = self.__read(old_child).set(
                            "parent_ref", WeakReference()
                        )
                        stores = stores._set(old_child, child_store)

                    # noinspection PyTypeChecker
                    for new_child in change.new_children:
                        children = children.add(new_child)
                        child_store = self.__read(new_child).set(
                            "parent_ref", WeakReference(obj)
                        )
                        if new_child in new_children_last_parent_history_updates:
                            child_store = child_store.set(
                                "last_parent_history_ref", WeakReference(history)
                            )
                        stores = stores._set(new_child, child_store)
                    store = store.set("children", children)
                stores = stores._set(obj, store)

                # History propagation.
                for adopter in filtered_history_adopters:
                    try:
                        adopter_store = stores[adopter]
                    except KeyError:
                        adopter_store = self.__read(adopter)
                    adopter_store = adopter_store.set(
                        "history_provider_ref", WeakReference(obj)
                    )
                    stores = stores._set(adopter, adopter_store)

                # Upstream data changes.
                if data is not old_data:
                    child_data = data
                    child = None
                    for i, parent in enumerate(hierarchy):
                        if i == 0:
                            child = parent
                            continue
                        assert child is not None

                        location = single_locations[i]
                        relationship = cast(
                            "Relationship", parent._get_relationship(location)
                        )
                        if not relationship.data:
                            break

                        assert data is not None
                        assert child_data is not None

                        data_location = single_data_locations[i]

                        parent_old_store = self.__read(parent)
                        parent_new_store = type(
                            parent
                        ).__functions__.replace_child_data(
                            parent_old_store,
                            child,
                            data_location,
                            child_data,
                        )
                        if parent_new_store is parent_old_store:
                            break
                        stores = stores._set(parent, parent_new_store)

                        child = parent
                        child_data = parent_new_store.data

                # Commit!
                commit = Commit(actions=actions, stores=stores)
                self.__commits.append(commit)

                # Push change to history.
                if (
                    history is not None
                    and history_provider is not None
                    and not obj._initializing
                    and not history_provider._initializing
                ):
                    history.__push_change__(change)

                # Post phase.
                for action in actions:
                    self.__react(action.receiver, action, Phase.POST)

                # Exit history atomic batch.
                if history is not None and atomic_batch_change is not None:
                    history.__exit_batch__(atomic_batch_change)

    def __update_metadata(
        self,
        obj,  # type: BaseObject
        update,  # type: Mapping[str, Any]
    ):
        # type: (...) -> None
        """
        Update metadata.

        :param obj: Object.
        :param update: Metadata update.
        """
        # Store changes.
        try:
            stores = self.__commits[-1].stores
        except IndexError:
            stores = InteractiveDictData()
        store = self.__read(obj)
        old_metadata = store.metadata
        store = store.update({"metadata": old_metadata.update(update)})
        stores = stores._set(obj, store)

        # Commit!
        commit = Commit(actions=(), stores=stores)
        self.__commits.append(commit)

    def __revert(self, index):
        # type: (int) -> None
        """
        Revert changes to a particular index.

        :param index: Index.
        """
        del self.__commits[index:]

    def __push(self):
        # type: () -> None
        """Push and merge changes to permanent storage."""
        if self.__commits:
            commits = self.__commits
            self.__commits = []

            action_exception_infos = []  # type: List[ActionObserverExceptionData]

            def ingest_action_exception_infos(result):
                # type: (Tuple[ObserverExceptionInfo, ...]) -> None
                """
                Ingest exception information.

                :param result: Exception information from subject-observers.
                """
                for exception_info in result:
                    internal_observer = cast(
                        "InternalObserver", exception_info.observer
                    )
                    action_observer = internal_observer.action_observer_ref()
                    if action_observer is not None:
                        from objetto._observers import ActionObserverExceptionData

                        action_exception_info = ActionObserverExceptionData(
                            observer=action_observer,
                            action=cast("Action", exception_info.payload[0]),
                            phase=cast("Phase", exception_info.payload[1]),
                            exception_type=exception_info.exception_type,
                            exception=exception_info.exception,
                            traceback=exception_info.traceback,
                        )
                        action_exception_infos.append(action_exception_info)

            for commit in commits:
                if type(commit) is BatchCommit:
                    for action in commit.actions:
                        phase = commit.phase  # type: ignore
                        ingest_action_exception_infos(
                            action.receiver.__.subject.send(
                                action, cast("Phase", phase)
                            )
                        )
                else:
                    for action in commit.actions:
                        ingest_action_exception_infos(
                            action.receiver.__.subject.send(action, Phase.PRE)
                        )

                    self.__storage.update(commit.stores)

                    for action in commit.actions:
                        ingest_action_exception_infos(
                            action.receiver.__.subject.send(action, Phase.POST)
                        )

            if action_exception_infos:
                raise ActionObserversFailedError(
                    "external observers raised exceptions (see tracebacks below)",
                    tuple(action_exception_infos),
                )

    @contextmanager
    def __hierarchy_context(self, obj):
        # type: (BaseObject) -> Iterator[List[BaseObject]]
        """
        Context manager that locks and caches an object's upper hierarchy.

        :param obj: Object.
        :return: Cached upper hierarchy (starting with the object itself).
        """
        hierarchy = []  # type: List[BaseObject]
        parent = obj  # type: Optional[BaseObject]
        while parent is not None:
            self.__busy_hierarchy[parent] += 1
            hierarchy.append(parent)
            # noinspection PyCallingNonCallable
            parent = self.__read(parent).parent_ref()
        try:
            yield hierarchy
        finally:
            for parent in hierarchy:
                self.__busy_hierarchy[parent] -= 1
                if not self.__busy_hierarchy[parent]:
                    del self.__busy_hierarchy[parent]

    @contextmanager
    def __new_children_context(self, new_children):
        # type: (AbstractSet[BaseObject]) -> Iterator
        """
        Context manager that locks parenting for new children.

        :param new_children: New children.
        """
        for new_child in new_children:
            self.__busy_hierarchy[new_child] += 1
        try:
            yield
        finally:
            for new_child in new_children:
                self.__busy_hierarchy[new_child] -= 1
                if not self.__busy_hierarchy[new_child]:
                    del self.__busy_hierarchy[new_child]

    @staticmethod
    def __react(obj, action, phase):
        # type: (BaseObject, Action, Phase) -> None
        """
        Run object's reactions.

        :param obj: Object.
        :param action: Action.
        :param phase: Phase.
        """
        for reaction in type(obj)._reactions:
            reaction(obj, action, phase)

    def init_object(self, obj):
        # type: (BaseObject) -> None
        """
        Initialize object.

        :param obj: Object.
        """
        with self.write_context():
            try:
                stores = self.__commits[-1].stores
            except IndexError:
                stores = InteractiveDictData()

            if obj in stores or obj in self.__storage:
                error = "object {} can't be initialized more than once".format(obj)
                raise RuntimeError(error)

            cls = type(obj)  # type: Type[BaseObject]
            kwargs = {}  # type: Dict[str, Any]

            # History object.
            history_descriptor = cls._history_descriptor
            if history_descriptor is not None:
                app = self.__app_ref()
                assert app is not None

                if self.__history_cls is None:
                    from ._history import HistoryObject

                    self.__history_cls = HistoryObject

                kwargs.update(
                    history_provider_ref=WeakReference(obj),
                    history=self.__history_cls(app, size=history_descriptor.size),
                )

            # State.
            state = cls._state_factory()

            # Data.
            data_type = cls.Data
            if data_type is not None:
                data = data_type.__make__()  # type: ignore
            else:
                data = None

            # Commit!
            stores = stores._set(obj, Store(state=state, data=data, **kwargs))
            commit = Commit(stores=stores)
            self.__commits.append(commit)

    @contextmanager
    def read_context(self, obj=None):
        # type: (Optional[BaseObject]) -> Iterator[ReadFunction]
        """
        Read context manager.

        :param obj: Object.
        :return: Read handle function.
        """
        with self.__lock:
            topmost = not self.__reading
            self.__reading.append(obj)

            def read():
                # type: () -> Store
                """Read object store."""
                assert obj is not None
                return self.__read(obj)

            try:
                yield read
            finally:
                self.__reading.pop()
                if topmost:
                    assert not self.__reading

    @contextmanager
    def write_context(
        self,
        obj=None,  # type: Optional[BaseObject]
    ):
        # type: (...) -> Iterator[Tuple[ReadFunction, WriteFunction]]
        """
        Write context manager.

        :param obj: Object.
        :return: Read and write handle functions.
        """
        with self.__lock:
            if self.__reading:
                error = "can't enter a 'write' context while in a 'read' context"
                raise RuntimeError(error)
            topmost = not self.__writing
            index = len(self.__commits)
            self.__writing.append(obj)

            def read():
                # type: () -> Store
                """Read object store."""
                assert obj is not None
                return self.__read(obj)

            def write(
                state,  # type: Any
                data,  # type: BaseData
                metadata,  # type: Mapping[str, Any]
                child_counter,  # type: Counter[BaseObject]
                change,  # type: BaseAtomicChange
            ):
                # type: (...) -> None
                """Write changes to object."""
                assert obj is not None
                if obj in self.__busy_writing:
                    error_ = "reaction cycle detected on {}".format(obj)
                    raise RuntimeError(error_)
                self.__busy_writing.add(obj)
                try:
                    self.__write(obj, state, data, metadata, child_counter, change)
                except RejectChangeException as e_:
                    self.__busy_writing.remove(obj)
                    if e_.change is not change:
                        raise
                    self.__revert(index)
                    e_.callback()
                except Exception:
                    self.__busy_writing.remove(obj)
                    raise
                else:
                    self.__busy_writing.remove(obj)

            try:
                yield read, write
            except Exception as e:
                self.__revert(index)
                if not topmost or type(e) is not TemporaryContextException:
                    raise
            else:
                if topmost:
                    with self.read_context():
                        self.__push()
            finally:
                self.__writing.pop()
                if topmost:
                    assert not self.__busy_hierarchy
                    assert not self.__busy_writing
                    assert not self.__commits
                    assert not self.__writing

    @contextmanager
    def update_metadata_context(
        self,
        obj,  # type: BaseObject
    ):
        # type: (...) -> Iterator[Tuple[ReadMetadataFunction, UpdateMetadataFunction]]
        """
        Update metadata context manager.

        :param obj: Object.
        :return: Read metadata and write metadata handle functions.
        """
        with self.write_context():

            def read_metadata():
                # type: () -> InteractiveDictData
                """Read metadata."""
                return self.__read(obj).metadata

            def update_metadata(
                update,  # type: Mapping[str, Any]
            ):
                # type: (...) -> None
                """Update metadata."""
                self.__update_metadata(obj, update)

            yield read_metadata, update_metadata

    @contextmanager
    def batch_context(self, obj, change):
        # type: (BaseObject, Batch) -> Iterator
        """
        Batch change context.

        :param obj: Object.
        :param change: Batch change.
        """
        with self.write_context():
            index = len(self.__commits)

            def _get_stores():
                try:
                    return self.__commits[-1].stores
                except IndexError:
                    return {}

            try:
                with self.__hierarchy_context(obj) as hierarchy:
                    assert hierarchy[0] is obj

                    # Get history.
                    history, history_provider = self.__read_history(obj)

                    # Gather actions.
                    child = None  # type: Optional[BaseObject]
                    single_locations = []  # type: List[Any]
                    all_locations = []  # type: List[List[Any]]
                    actions = []  # type: List[Action]
                    for i, parent in enumerate(hierarchy):
                        if i == 0:
                            location = None
                            locations = []
                        else:
                            assert child is not None
                            location = parent._locate(child)
                            locations = [location] + all_locations[-1]

                        single_locations.append(location)
                        all_locations.append(locations)

                        action = Action(
                            sender=obj,
                            receiver=parent,
                            locations=locations,
                            change=change,
                        )
                        assert action.sender is obj
                        assert action.receiver is parent

                        actions.append(action)
                        child = parent

                    # Commit Pre.
                    pre_commit = BatchCommit(
                        actions=actions, stores=_get_stores(), phase=Phase.PRE
                    )
                    self.__commits.append(pre_commit)

                    # History Pre.
                    if (
                        history is not None
                        and history_provider is not None
                        and not obj._initializing
                        and not history_provider._initializing
                    ):
                        history.__enter_batch__(change)

                    # Pre.
                    for action in actions:
                        self.__react(action.receiver, action, Phase.PRE)

                    yield change

                    # History Post.
                    if (
                        history is not None
                        and history_provider is not None
                        and not obj._initializing
                        and not history_provider._initializing
                    ):
                        history.__exit_batch__(change)

                    # Post.
                    for action in actions:
                        self.__react(action.receiver, action, Phase.POST)

                    # Commit Post.
                    post_commit = BatchCommit(
                        actions=actions, stores=_get_stores(), phase=Phase.POST
                    )
                    self.__commits.append(post_commit)

            # Catch rejection.
            except RejectChangeException as e:
                self.__revert(index)
                if e.change is not change:
                    raise
                e.callback()

    def init_root_objs(self):
        # type: () -> None
        """Initialize root objects."""
        app = self.__app_ref()
        assert app is not None
        assert not self.__roots
        roots = type(app)._roots
        if roots:
            with self.write_context():
                sorted_roots = sorted(
                    itervalues(roots), key=lambda r: (r.priority is None, r.priority)
                )
                for root in sorted_roots:
                    root_obj = root.obj_type(app, **root.kwargs)
                    self.__roots[root] = root_obj
                    root_obj.__.set_root()

    def get_root_obj(self, root):
        # type: (ApplicationRoot) -> BaseObject
        """
        Get root object.

        :param root: Application root descriptor.
        :return: Root object.
        """
        return self.__roots[root]


class ApplicationMeta(BaseMeta):
    """Metaclass for `Application`."""

    __roots = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ApplicationMeta, Mapping[str, ApplicationRoot]]
    __root_names = WeakKeyDictionary(
        {}
    )  # type: MutableMapping[ApplicationMeta, Mapping[ApplicationRoot, str]]

    def __init__(cls, name, bases, dct):
        # type: (str, Tuple[Type, ...], Dict[str, Any]) -> None
        super(ApplicationMeta, cls).__init__(name, bases, dct)

        # Store roots.
        roots = {}
        for base in reversed(getmro(cls)):
            for member_name, member in iteritems(base.__dict__):
                if isinstance(member, ApplicationRoot):
                    roots[member_name] = member
                elif member_name in roots:
                    del roots[member_name]

        # Store root names.
        root_names = {}
        for root_name, root in iteritems(roots):
            root_names[root] = root_name

        type(cls).__roots[cls] = DictState(roots)
        type(cls).__root_names[cls] = DictState(root_names)

    @property
    def _roots(cls):
        # type: () -> Mapping[str, ApplicationRoot]
        """Attributes mapped by name."""
        return type(cls).__roots[cls]

    @property
    def _root_names(cls):
        # type: () -> Mapping[Any, str]
        """Names mapped by root."""
        return type(cls).__root_names[cls]


class Application(with_metaclass(ApplicationMeta, Base)):
    """
    Application.

    Inherits from:
      - :class:`objetto.bases.Base`

    Features:
      - Manages multiple :class:`objetto.bases.BaseObject` under the same hierarchy.
      - Offers contexts for reading/writing/batch.
      - Reverts changes when an error occurs.
      - Manages :class:`objetto.objects.Action` propagation, internally and externally.

    When initializing an :class:`objetto.objects.BaseObject`, you have to pass an
    :class:`objetto.applications.Application` as its first parameter.

    .. code:: python

        >>> from objetto import Application, Object

        >>> app = Application()
        >>> obj = Object(app)  # pass application as first parameter
        >>> obj.app is app  # access it through the 'app' property
        True
    """

    __slots__ = ("__weakref__", "__")

    def __init__(self):
        self.__ = ApplicationInternals(self)
        self.__.init_root_objs()

    @final
    @contextmanager
    def read_context(self):
        # type: () -> Iterator
        """
        Read context.

        :return: Context manager.
        :rtype: contextlib.AbstractContextManager
        """
        with self.__.read_context():
            yield

    @final
    @contextmanager
    def write_context(self):
        # type: () -> Iterator
        """
        Write context.

        :return: Context manager.
        :rtype: contextlib.AbstractContextManager
        """
        with self.__.write_context():
            yield

    @final
    @contextmanager
    def temporary_context(self):
        # type: () -> Iterator
        """
        Temporary write context.

        :return: Context manager.
        :rtype: contextlib.AbstractContextManager
        """
        with self.__.write_context():
            try:
                yield
            except Exception:
                raise
            else:
                raise TemporaryContextException()



# -*- coding: utf-8 -*-
"""Import and run factory functions/classes."""

from typing import TYPE_CHECKING

from six import string_types

from .lazy_import import decorate_path, import_path

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Mapping, Optional, Union

    LazyFactory = Optional[Union[Callable, str]]
else:
    LazyFactory = None

__all__ = [
    "get_factory_name",
    "format_factory",
    "LazyFactory",
    "import_factory",
    "run_factory",
]


def get_factory_name(factory):
    # type: (LazyFactory) -> str
    """
    Get factory name without importing lazy paths.

    .. code:: python

        >>> from objetto.utils.factoring import get_factory_name

        >>> get_factory_name("module.submodule|Class.method")
        'method'
        >>> get_factory_name(int)
        'int'

    :param factory: Factory.
    :type factory: str or function or collections.abc.Callable or None

    :return: Factory name.
    :rtype: str
    """
    if isinstance(factory, string_types):
        return factory.split("|")[-1].split(".")[-1]
    elif factory is None:
        return "None"
    else:
        return factory.__name__


def format_factory(factory, module=None):
    # type: (LazyFactory, Optional[str]) -> LazyFactory
    """
    Check and format factory by adding a module path if applicable.

    .. code:: python

        >>> from objetto.utils.factoring import format_factory

        >>> format_factory("Class.method", module="module.submodule")
        'module.submodule|Class.method'
        >>> format_factory(int, module="module.submodule").__name__
        'int'

    :param factory: Factory.
    :type factory: str or function or collections.abc.Callable or None

    :param module: Module to prefix lazy paths without one.
    :type module: str or None

    :return: Formatted factory.
    :rtype: str or function or collections.abc.Callable or None

    :raises ValueError: Invalid import path to factory.
    :raises TypeError: Did not provide valid factory.
    """
    if factory is None or callable(factory):
        return factory
    elif isinstance(factory, string_types):
        return decorate_path(factory, module=module)
    else:
        raise TypeError(type(factory).__name__)


def import_factory(factory):
    # type: (LazyFactory) -> Optional[Callable]
    """
    Import factory from a path.

    .. code:: python

        >>> from re import match
        >>> from objetto.utils.factoring import import_factory

        >>> import_factory("abc|abstractmethod")
        <function abstractmethod at ...>
        >>> import_factory(None)  # returns None
        >>> import_factory(match)
        <function match at ...>

    :param factory: Lazy factory.
    :type factory: str or function or collections.abc.Callable or None

    :return: Imported factory.
    :rtype: function or collections.abc.Callable or None

    :raises TypeError: Invalid factory.
    """
    if factory is None or callable(factory):
        return factory
    elif isinstance(factory, string_types):
        return import_path(factory)
    else:
        raise TypeError(type(factory).__name__)


def run_factory(
    factory,  # type: LazyFactory
    args=None,  # type: Optional[Iterable[Any]]
    kwargs=None,  # type: Optional[Mapping[str, Any]]
):
    # type: (...) -> Any
    """
    Import and run factory.

    .. code:: python

        >>> from objetto.utils.factoring import run_factory

        >>> bool(run_factory("re|match", (r"^[a-z]+$", "abc")))
        True

    :param factory: Lazy factory.
    :type factory: str or function or collections.abc.Callable or None

    :param args: Arguments to be passed to the factory function.

    :param kwargs: Keyword arguments to be passed to the factory function.

    :return: Result from factory.
    """
    factory = import_factory(factory)
    if factory is None:
        return None
    else:
        return factory(*(args or ()), **(kwargs or {}))



# -*- coding: utf-8 -*-

import pytest

from objetto.utils.factoring import (
    format_factory,
    get_factory_name,
    import_factory,
    run_factory,
)


def test_format_factory():
    assert format_factory(None) is None
    assert format_factory("Class.method", module="module.submodule") == (
        "module.submodule|Class.method"
    )
    assert format_factory("function", module="module.submodule") == (
        "module.submodule|function"
    )
    assert format_factory("module.submodule|Class.method", module="") == (
        "module.submodule|Class.method"
    )
    assert format_factory("module.submodule|function", module="") == (
        "module.submodule|function"
    )
    assert format_factory(int) is int
    with pytest.raises(TypeError):
        params = (3,)
        format_factory(*params)


def test_get_factory_name():
    assert get_factory_name("Class.method") == "method"
    assert get_factory_name("function") == "function"
    assert get_factory_name("module.submodule|Class.method") == "method"
    assert get_factory_name("module.submodule|function") == "function"
    assert get_factory_name(None) == "None"
    assert get_factory_name(int) == "int"
    assert get_factory_name(float) == "float"


def test_import_factory():
    from abc import abstractmethod
    from re import match

    assert import_factory("abc|abstractmethod") is abstractmethod
    assert import_factory(None) is None
    assert import_factory(match) is match


def test_run_factory():
    def my_factory(*args, **kwargs):
        assert args == (1, 2, 3)
        assert kwargs == {"a": 1, "b": 2, "c": 3}

    assert bool(run_factory("re|match", (r"^[a-z]+$", "abc"))) is True
    run_factory(my_factory, (1, 2, 3), dict(a=1, b=2, c=3))


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
"""
An :class:`objetto.applications.Application` oversees all
:class:`objetto.bases.BaseObject` objects that are meant to work together.
It provides different contexts for managing and keeping track of their changes.

Objects that are part of different applications see each other as regular values and
can never be part of the same hierarchy.

An application can have root objects defined by :func:`objetto.applications.root`,
which are always available at the top of the hierarchy, and cannot be parented under
other objects.
"""

from typing import TYPE_CHECKING

from ._applications import BO, Application, ApplicationRoot

if TYPE_CHECKING:
    from typing import Any, Optional, Type

__all__ = ["Application", "root"]


def root(obj_type, priority=None, **kwargs):
    # type: (Type[BO], Optional[int], Any) -> ApplicationRoot[BO]
    """
    Describes a root object that gets initialized with the application.

    .. code:: python

        >>> from objetto import Application, Object, attribute, root

        >>> class MyObject(Object):
        ...     name = attribute(str)
        ...
        >>> class MyApplication(Application):  # subclass for adding roots
        ...     root_a = root(MyObject, name="Root A")  # describe root
        ...     root_b = root(MyObject, name="Root B")  # describe root
        ...
        >>> app = MyApplication()  # instantiate the application
        >>> app.root_a  # access a root object
        MyObject(name='Root A')

    :param obj_type: Object type.
    :type obj_type: type[objetto.bases.BaseObject]

    :param priority: Initialization priority.
    :type priority: int or None

    :param kwargs: Keyword arguments to be passed to the object's `__init__`.

    :return: Application root descriptor.
    :rtype: objetto.applications.ApplicationRoot

    :raises ValueError: Used reserved keyword argument.
    :raises TypeError: Invalid object type.
    """
    return ApplicationRoot(obj_type, priority=priority, **kwargs)



# -*- coding: utf-8 -*-
"""Factories."""

from abc import abstractmethod
from re import compile as re_compile
from re import match as re_match
from re import sub as re_sub
from typing import TYPE_CHECKING, cast

from six import string_types

from ._bases import Base
from .utils.factoring import format_factory, run_factory
from .utils.reraise_context import ReraiseContext
from .utils.type_checking import assert_is_callable, assert_is_instance

if TYPE_CHECKING:
    from re import Pattern
    from typing import Any, Callable, Iterable, Optional, Tuple, Union

    from .utils.factoring import LazyFactory

    FactoryType = Union[Callable[[Any, Any], Any], "BaseFactory"]

__all__ = [
    "BaseFactory",
    "MultiFactory",
    "Integer",
    "FloatingPoint",
    "RegexMatch",
    "RegexSub",
    "String",
    "Curated",
    "Boolean",
]


class BaseFactory(Base):
    """
    Base callable factory object.

    Inherits from:
      - :class:`objetto.bases.Base`

    Inherited By:
      - :class:`objetto.factories.MultiFactory`
      - :class:`objetto.factories.Integer`
      - :class:`objetto.factories.FloatingPoint`
      - :class:`objetto.factories.RegexMatch`
      - :class:`objetto.factories.RegexSub`
      - :class:`objetto.factories.String`
      - :class:`objetto.factories.Curated`
      - :class:`objetto.factories.Boolean`
    """

    __slots__ = ("__factories",)

    @abstractmethod
    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.

        :param kwargs: Keyword arguments.

        :return: Output value.
        """
        raise NotImplementedError()

    def __add__(self, other):
        # type: (Union[BaseFactory, FactoryType, LazyFactory]) -> MultiFactory
        """
        Add with another factory.

        :param other: Another factory.
        :type other: objetto.bases.BaseFactoryor str or collections.abc.Callable or None

        :return: Multi factory with added factories.
        :rtype: objetto.factories.MultiFactory
        """
        with ReraiseContext(TypeError, "adding factories together"):
            if not isinstance(other, string_types):
                assert_is_callable(other)
        return MultiFactory((self, other))


class MultiFactory(BaseFactory):
    """
    Adds multiple factories together.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`

    :param factories: Factories to be added together.
    :type factories: collections.abc.Iterable[objetto.bases.BaseFactory or \
 collections.abc.Callable or function or str or None]
    """

    __slots__ = ("__factories",)

    def __init__(
        self,
        factories,  # type: Iterable[Union[BaseFactory, FactoryType, LazyFactory]]
        module=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        self.__factories = tuple(
            format_factory(r, module=module) for r in factories if r is not None
        )

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.

        :param kwargs: Keyword arguments.

        :return: Output value.
        """
        for factory in self.factories:
            value = run_factory(factory, args=(value,), kwargs=kwargs)
        return value

    def __add__(self, other):
        # type: (Union[BaseFactory, FactoryType, LazyFactory]) -> MultiFactory
        """
        Add with another factory.

        :param other: Another factory.
        :type other: objetto.bases.BaseFactoryor str or collections.abc.Callable or None

        :return: Multi factory with added factories.
        :rtype: objetto.factories.MultiFactory
        """
        if isinstance(other, MultiFactory):
            return MultiFactory(self.__factories + other.__factories)
        else:
            return MultiFactory(self.__factories + (other,))

    @property
    def factories(self):
        # type: () -> Tuple[Union[BaseFactory, FactoryType, LazyFactory], ...]
        """
        Factories.

        :return: tuple[objetto.bases.BaseFactory or \
 collections.abc.Callable or function or str]
        """
        return self.__factories


class Integer(BaseFactory):
    """
    Integer factory.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`

    :param minimum: Minimum.
    :type minimum: int or None

    :param maximum: Maximum.
    :type maximum: int or None

    :param clamp_minimum: Whether to clamp to minimum value instead of erroring.
    :type clamp_minimum: bool

    :param clamp_maximum: Whether to clamp to maximum value instead of erroring.
    :type clamp_maximum: bool

    :param accepts_none: Whether to accept None as a value.
    :type accepts_none: bool

    :raises ValueError: Invalid parameter value.
    """

    __slots__ = (
        "__minimum",
        "__maximum",
        "__clamp_minimum",
        "__clamp_maximum",
        "__accepts_none",
    )

    def __init__(
        self,
        minimum=None,  # type: Optional[int]
        maximum=None,  # type: Optional[int]
        clamp_minimum=False,  # type: bool
        clamp_maximum=False,  # type: bool
        accepts_none=False,  # type: bool
    ):
        # type: (...) -> None
        if clamp_minimum and minimum is None:
            error = "parameter 'clamp_minimum' was set to True, but no minimum provided"
            raise ValueError(error)
        if clamp_maximum and maximum is None:
            error = "parameter 'clamp_maximum' was set to True, but no maximum provided"
            raise ValueError(error)

        if minimum is not None:
            minimum = int(minimum)
        if maximum is not None:
            maximum = int(maximum)

        self.__minimum = minimum
        self.__maximum = maximum
        self.__clamp_minimum = bool(clamp_minimum)
        self.__clamp_maximum = bool(clamp_maximum)
        self.__accepts_none = bool(accepts_none)

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.
        :type value: int or None

        :param kwargs: Keyword arguments.

        :return: Output value.
        :rtype: int or None

        :raises ValueError: Value out of bounds.
        """
        if value is None and self.__accepts_none:
            return value
        value = int(value)
        if self.minimum is not None and value < self.minimum:
            if self.clamp_minimum:
                value = self.minimum
            else:
                error_msg = "minimum value is {}, got {}".format(self.minimum, value)
                raise ValueError(error_msg)
        if self.maximum is not None and value > self.maximum:
            if self.clamp_maximum:
                value = self.maximum
            else:
                error_msg = "maximum value is {}, got {}".format(self.minimum, value)
                raise ValueError(error_msg)
        return value

    @property
    def minimum(self):
        # type: () -> Optional[int]
        """
        Minimum.

        :rtype: int or None
        """
        return self.__minimum

    @property
    def maximum(self):
        # type: () -> Optional[int]
        """
        Maximum.

        :rtype: int or None
        """
        return self.__maximum

    @property
    def clamp_minimum(self):
        # type: () -> bool
        """
        Whether to clamp to minimum value instead of erroring.

        :rtype: bool
        """
        return self.__clamp_minimum

    @property
    def clamp_maximum(self):
        # type: () -> bool
        """
        Whether to clamp to maximum value instead of erroring.

        :rtype: bool
        """
        return self.__clamp_maximum

    @property
    def accepts_none(self):
        # type: () -> bool
        """
        Whether to accept None as a value.

        :rtype: bool
        """
        return self.__accepts_none


class FloatingPoint(BaseFactory):
    """
    Floating point factory.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`

    :param minimum: Minimum.
    :type minimum: float or None

    :param maximum: Maximum.
    :type maximum: float or None

    :param clamp_minimum: Whether to clamp to minimum value instead of erroring.
    :type clamp_minimum: bool

    :param clamp_maximum: Whether to clamp to maximum value instead of erroring.
    :type clamp_maximum: bool

    :param accepts_none: Whether to accept None as a value.
    :type accepts_none: bool

    :raises ValueError: Invalid parameter value.
    """

    __slots__ = (
        "__minimum",
        "__maximum",
        "__clamp_minimum",
        "__clamp_maximum",
        "__accepts_none",
    )

    def __init__(
        self,
        minimum=None,  # type: Optional[float]
        maximum=None,  # type: Optional[float]
        clamp_minimum=False,  # type: bool
        clamp_maximum=False,  # type: bool
        accepts_none=False,  # type: bool
    ):
        # type: (...) -> None
        if clamp_minimum and minimum is None:
            error = "parameter 'clamp_minimum' was set to True, but no minimum provided"
            raise ValueError(error)
        if clamp_maximum and maximum is None:
            error = "parameter 'clamp_maximum' was set to True, but no maximum provided"
            raise ValueError(error)

        if minimum is not None:
            minimum = float(minimum)
        if maximum is not None:
            maximum = float(maximum)

        self.__minimum = minimum
        self.__maximum = maximum
        self.__clamp_minimum = bool(clamp_minimum)
        self.__clamp_maximum = bool(clamp_maximum)
        self.__accepts_none = bool(accepts_none)

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.
        :type value: float or None

        :param kwargs: Keyword arguments.

        :return: Output value.
        :rtype: float or None

        :raises ValueError: Value out of bounds.
        """
        if value is None and self.__accepts_none:
            return value
        value = float(value)
        if self.minimum is not None and value < self.minimum:
            if self.clamp_minimum:
                value = self.minimum
            else:
                error_msg = "minimum value is {}, got {}".format(self.minimum, value)
                raise ValueError(error_msg)
        if self.maximum is not None and value > self.maximum:
            if self.clamp_maximum:
                value = self.maximum
            else:
                error_msg = "maximum value is {}, got {}".format(self.minimum, value)
                raise ValueError(error_msg)
        return value

    @property
    def minimum(self):
        # type: () -> Optional[float]
        """
        Minimum.

        :rtype: float or None
        """
        return self.__minimum

    @property
    def maximum(self):
        # type: () -> Optional[float]
        """
        Maximum.

        :rtype: float or None
        """
        return self.__maximum

    @property
    def clamp_minimum(self):
        # type: () -> bool
        """
        Whether to clamp to minimum value instead of erroring.

        :rtype: bool
        """
        return self.__clamp_minimum

    @property
    def clamp_maximum(self):
        # type: () -> bool
        """
        Whether to clamp to maximum value instead of erroring.

        :rtype: bool
        """
        return self.__clamp_maximum

    @property
    def accepts_none(self):
        # type: () -> bool
        """
        Whether to accept None as a value.

        :rtype: bool
        """
        return self.__accepts_none


class String(BaseFactory):
    """
    String factory.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`

    :param accepts_none: Whether to accept None.
    :type accepts_none: bool
    """

    __slots__ = ("__accepts_none",)

    def __init__(self, accepts_none=False):
        # type: (bool) -> None
        self.__accepts_none = bool(accepts_none)

    def __call__(self, value, **kwargs):
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.
        :type value: str or None

        :param kwargs: Keyword arguments.

        :return: Output value.
        :rtype: str or None

        :raises ValueError: Value out of bounds.
        """
        if value is None and self.__accepts_none:
            return value
        if not isinstance(value, string_types):
            return str(value)
        else:
            return value

    @property
    def accepts_none(self):
        # type: () -> bool
        """
        Whether to accept None.

        :rtype: bool
        """
        return self.__accepts_none


class RegexMatch(String):
    """
    Regex match check factory.

    Inherits from:
      - :class:`objetto.factories.String`

    :param pattern: Regex pattern.
    :type pattern: str

    :param accepts_none: Whether to accept None.
    :type accepts_none: bool
    """

    __slots__ = ("__pattern", "__compiled_pattern")

    def __init__(self, pattern, accepts_none=False):
        # type: ("Union[str, bytes]", bool) -> None
        super(RegexMatch, self).__init__(accepts_none=accepts_none)
        self.__pattern = cast("Union[str, bytes]", pattern)
        self.__compiled_pattern = re_compile(pattern)  # type: Pattern

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.
        :type value: str or None

        :param kwargs: Keyword arguments.

        :return: Output value.
        :rtype: str or None

        :raises ValueError: Value out of bounds.
        """
        value = super(RegexMatch, self).__call__(value, **kwargs)
        if value is None:
            return value
        if not re_match(self.compiled_pattern, value):
            error_msg = "'{}' does not match regex pattern '{}'".format(
                value, str(self.pattern)
            )
            raise ValueError(error_msg)
        return value

    @property
    def pattern(self):
        # type: () -> "Union[str, bytes]"
        """
        Regex pattern.

        :rtype: str
        """
        return self.__pattern

    @property
    def compiled_pattern(self):
        # type: () -> Pattern
        """
        Compiled regex pattern.

        :rtype: re.Pattern
        """
        return self.__compiled_pattern


class RegexSub(String):
    """
    Regex substitution factory.

    Inherits from:
      - :class:`objetto.factories.String`

    :param pattern: Regex pattern.
    :type pattern: str

    :param repl: Substitution.
    :type repl: str

    :param accepts_none: Whether to accept None.
    :type accepts_none: bool
    """

    __slots__ = ("__pattern", "__compiled_pattern", "__repl", "__accepts_none")

    def __init__(self, pattern, repl, accepts_none=False):
        # type: (str, str, bool) -> None
        super(RegexSub, self).__init__(accepts_none=accepts_none)
        with ReraiseContext(TypeError, "'pattern' parameter"):
            assert_is_instance(pattern, string_types)
        with ReraiseContext(TypeError, "'repl' parameter"):
            assert_is_instance(repl, string_types)
        self.__pattern = pattern
        self.__compiled_pattern = re_compile(pattern)
        self.__repl = repl

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.
        :type value: str or None

        :param kwargs: Keyword arguments.

        :return: Output value.
        :rtype: str or None

        :raises ValueError: Value out of bounds.
        """
        value = super(RegexSub, self).__call__(value, **kwargs)
        return re_sub(self.compiled_pattern, self.repl, value)

    @property
    def pattern(self):
        # type: () -> "Union[str, bytes]"
        """
        Regex pattern.

        :rtype: str
        """
        return self.__pattern

    @property
    def compiled_pattern(self):
        # type: () -> Pattern
        """
        Compiled regex pattern.

        :rtype: re.Pattern
        """
        return self.__compiled_pattern

    @property
    def repl(self):
        """
        Substitution.

        :rtype: str
        """
        return self.__repl


class Curated(BaseFactory):
    """
    Curated values factory.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`

    :param values: Accepted values.
    """

    __slots__ = ("__values",)

    def __init__(self, *values):
        # type: (Any) -> None
        self.__values = values

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> Any
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.

        :param kwargs: Keyword arguments.

        :return: Output value.
        """
        if value not in self.values:
            error = "expected one of {}, got {}".format(self.values, repr(value))
            raise ValueError(error)
        return value

    @property
    def values(self):
        """
        Accepted values.

        :rtype: tuple
        """
        return self.__values


class Boolean(BaseFactory):
    """
    Boolean factory.

    Inherits from:
      - :class:`objetto.bases.BaseFactory`
    """

    __slots__ = ()

    def __call__(self, value, **kwargs):
        # type: (Any, Any) -> bool
        """
        Call with input value and optional keyword arguments.

        :param value: Input value.

        :param kwargs: Keyword arguments.

        :return: Output value (True or False).
        :rtype: bool
        """
        return bool(value)



# -*- coding: utf-8 -*-

from typing import TYPE_CHECKING

from ._bases import final
from ._changes import BaseAtomicChange, Batch
from ._exceptions import BaseObjettoException
from .factories import Integer
from .objects import (
    Object,
    attribute,
    protected_attribute_pair,
    protected_list_attribute_pair,
)

if TYPE_CHECKING:
    from typing import Any, Optional, TypeVar, Union

    from ._objects import (
        Attribute,
        DictObject,
        ListObject,
        MutableDictObject,
        MutableListObject,
        MutableSetObject,
        ProxyDictObject,
        ProxyListObject,
        ProxySetObject,
        SetObject,
    )

    T = TypeVar("T")  # Any type.
    KT = TypeVar("KT")  # Any key type.
    VT = TypeVar("VT")  # Any value type.

    MDA = Attribute[MutableDictObject[KT, VT]]
    MLA = Attribute[MutableListObject[T]]
    MSA = Attribute[MutableSetObject[T]]
    DA = Attribute[DictObject[KT, VT]]
    LA = Attribute[ListObject[T]]
    SA = Attribute[SetObject[T]]
    PDA = Attribute[ProxyDictObject[KT, VT]]
    PLA = Attribute[ProxyListObject[T]]
    PSA = Attribute[ProxySetObject[T]]
    CT = Union[BaseAtomicChange, "BatchChanges"]

__all__ = ["HistoryError", "BatchChanges", "HistoryObject"]


class HistoryError(BaseObjettoException):
    """
    History failed to execute.

    Inherits from:
      - :class:`objetto.bases.BaseObjettoException`
    """


# noinspection PyAbstractClass
@final
class BatchChanges(Object):
    """
    Batch changes.

    Inherits from:
      - :class:`objetto.objects.Object`
    """

    __slots__ = ()

    change = attribute(Batch, checked=False, changeable=False)  # type: Attribute[Batch]
    """
    Batch change with name and metadata.

    :type: objetto.changes.Batch
    """

    name = attribute(str, checked=False, changeable=False)  # type: Attribute[str]
    """
    The batch change name.

    :type: str
    """

    _changes, changes = protected_list_attribute_pair(
        (BaseAtomicChange, "BatchChanges"), subtypes=True, checked=False
    )  # type: PLA[CT], LA[CT]
    """
    Changes executed during the batch.

    :type: objetto.objects.ListObject[objetto.history.BatchChanges or \
 objetto.bases.BaseAtomicChange]
    """

    _closed, closed = protected_attribute_pair(
        bool, checked=False, default=False
    )  # type: Attribute[bool], Attribute[bool]
    """
    Whether the batch has already completed or is still running.

    :type: bool
    """

    def format_changes(self):
        # type: () -> str
        """
        Format changes into readable string.

        :return: Formatted changes.
        :rtype: str
        """
        with self.app.read_context():
            parts = []
            # noinspection PyTypeChecker
            for change in self.changes:
                if isinstance(change, BatchChanges):
                    parts.append(
                        "{} (Batch) ({})".format(
                            change.name, type(change.change.obj).__name__
                        )
                    )
                    for part in change.format_changes().split("\n"):
                        parts.append("  {}".format(part))
                else:
                    parts.append(
                        "{} ({})".format(change.name, type(change.obj).__name__)
                    )
            return "\n".join(parts)

    def __undo__(self, _):
        # type: (Any) -> None
        """Undo."""
        # noinspection PyTypeChecker
        for change in reversed(self.changes):
            change.__undo__(change)

    def __redo__(self, _):
        # type: (Any) -> None
        """Redo."""
        # noinspection PyTypeChecker
        for change in self.changes:
            change.__redo__(change)


# noinspection PyAbstractClass
@final
class HistoryObject(Object):
    """
    History object.

    Inherits from:
      - :class:`objetto.objects.Object`
    """

    __slots__ = ()

    size = attribute(
        (int, None),
        checked=False,
        default=None,
        factory=Integer(minimum=0, accepts_none=True),
        changeable=False,
    )  # type: Attribute[int]
    """
    How many changes to remember.

    :type: int
    """

    __executing, executing = protected_attribute_pair(
        bool, checked=False, default=False
    )  # type: Attribute[bool], Attribute[bool]
    """
    Whether the history is undoing or redoing.

    :type: bool
    """

    __undoing, undoing = protected_attribute_pair(
        bool, checked=False, default=False
    )  # type: Attribute[bool], Attribute[bool]
    """
    Whether the history is undoing.

    :type: bool
    """

    __redoing, redoing = protected_attribute_pair(
        bool, checked=False, default=False
    )  # type: Attribute[bool], Attribute[bool]
    """
    Whether the history is redoing.

    :type: bool
    """

    __index, index = protected_attribute_pair(
        int, checked=False, default=0
    )  # type: Attribute[int], Attribute[int]
    """
    The index of the current change.

    :type: int
    """

    __changes, changes = protected_list_attribute_pair(
        (BatchChanges, None),
        subtypes=True,
        checked=False,
        default=(None,),
    )  # type: PLA[Optional[BatchChanges]], LA[Optional[BatchChanges]]
    """
    List of batch changes. The first one is always `None`.

    :type: objetto.objects.ListObject[objetto.history.BatchChanges or None]
    """

    _current_batches, current_batches = protected_list_attribute_pair(
        BatchChanges,
        subtypes=False,
        checked=False,
        child=False,
    )  # type: PLA[BatchChanges], LA[BatchChanges]
    """
    Open batches.

    :type: objetto.objects.ListObject[objetto.history.BatchChanges]
    """

    def set_index(self, index):
        # type: (int) -> None
        """
        Undo/redo until we reach the desired index.

        :param index: Index.
        :type index: int

        :raise IndexError: Invalid index.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't set index while executing"
                raise HistoryError(error)
            if 0 <= index <= len(self.changes) - 1:
                if index > self.__index:
                    with self._batch_context("Multiple Redo"):
                        while index > self.__index:
                            self.redo()
                elif index < self.__index:
                    with self._batch_context("Multiple Undo"):
                        while index < self.__index:
                            self.undo()
            else:
                raise IndexError(index)

    def undo_all(self):
        # type: () -> None
        """
        Undo all.

        :raises HistoryError: Can't undo all while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't undo all while executing"
                raise HistoryError(error)
            with self._batch_context("Undo All"):
                while self.__index > 0:
                    self.undo()

    def redo_all(self):
        # type: () -> None
        """
        Redo all.

        :raises HistoryError: Can't redo all while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't redo all while executing"
                raise HistoryError(error)
            with self._batch_context("Redo All"):
                while self.__index < len(self.changes) - 1:
                    self.redo()

    # noinspection PyTypeChecker
    def redo(self):
        # type: () -> None
        """
        Redo.

        :raises HistoryError: Can't redo while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't redo while executing"
                raise HistoryError(error)
            if self.__index < len(self.changes) - 1:
                change = self.changes[self.__index + 1]
                assert change is not None
                with self._batch_context("Redo", change=change):
                    self.__executing = True
                    self.__redoing = True
                    try:
                        change.__redo__(change)
                    finally:
                        self.__executing = False
                        self.__redoing = False
                    self.__index += 1
            else:
                error = "can't redo any further"
                raise HistoryError(error)

    # noinspection PyTypeChecker
    def undo(self):
        # type: () -> None
        """
        Undo.

        :raises HistoryError: Can't undo while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't undo while executing"
                raise HistoryError(error)

            if self.__index > 0:
                change = self.changes[self.index]
                assert change is not None
                with self._batch_context("Undo", change=change):
                    self.__executing = True
                    self.__undoing = True
                    try:
                        change.__undo__(change)
                    finally:
                        self.__executing = False
                        self.__undoing = False
                    self.__index -= 1
            else:
                error = "can't undo any further"
                raise HistoryError(error)

    def flush(self):
        # type: () -> None
        """
        Flush all changes.

        :raises HistoryError: Can't flush while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't flush history while executing"
                raise HistoryError(error)
            if len(self.changes) > 1:
                with self._batch_context("Flush"):
                    # noinspection PyTypeChecker
                    self.__index = 0
                    del self.__changes[1:]

    def flush_redo(self):
        # type: () -> None
        """
        Flush changes ahead of the current index.

        :raises HistoryError: Can't flush while executing.
        """
        with self.app.write_context():
            if self.__executing:
                error = "can't flush history while executing"
                raise HistoryError(error)
            if len(self.changes) > 1:
                if self.__index < len(self.changes) - 1:
                    with self._batch_context("Flush Redo"):
                        del self.__changes[self.__index + 1 :]

    def in_batch(self):
        # type: () -> bool
        """
        Get whether history is currently in an open batch.

        :return: True if currently in an open batch.
        :rtype: bool

        :raises HistoryError: Can't check while executing.
        """
        with self.app.read_context():
            if self.__executing:
                error = "can't check if in a batch while executing"
                raise HistoryError(error)
            return bool(
                len(self.changes) > 1
                and isinstance(self.changes[-1], BatchChanges)
                and not self.changes[-1].closed
            )

    def format_changes(self):
        # type: () -> str
        """
        Format changes into readable string.

        :return: Formatted changes.
        :rtype: str
        """
        with self.app.read_context():
            parts = ["--- <-" if self.index == 0 else "---"]
            for i, change in enumerate(self.changes):
                if i == 0:
                    continue
                if isinstance(change, BatchChanges):
                    parts.append(
                        "{} (Batch) ({}) <-".format(
                            change.name, type(change.change.obj).__name__
                        )
                        if self.index == i
                        else "{} (Batch) ({})".format(
                            change.name, type(change.change.obj).__name__
                        )
                    )
                    for part in change.format_changes().split("\n"):
                        parts.append("  {}".format(part))
                elif change is not None:
                    parts.append(
                        "{} ({}) <-".format(change.name, type(change.obj).__name__)
                        if self.index == i
                        else "{} ({})".format(change.name, type(change.obj).__name__)
                    )
            return "\n".join(parts)

    def __enter_batch__(self, batch):
        # type: (Batch) -> None
        """
        Enter batch context.

        :param batch: Batch.
        """
        with self.app.write_context():
            if self.__executing:
                return
            with self._batch_context("Enter Batch", batch=batch):
                self.flush_redo()
                topmost = not self._current_batches
                batch_changes = BatchChanges(self.app, change=batch, name=batch.name)
                if topmost:
                    self.__changes.append(batch_changes)
                    if self.size is not None and len(self.changes) > self.size + 1:
                        del self.__changes[1:2]
                    else:
                        self.__index += 1
                else:
                    self._current_batches[-1]._changes.append(batch_changes)
                self._current_batches.append(batch_changes)

    def __push_change__(self, change):
        # type: (BaseAtomicChange) -> None
        """
        Push change to the current batch.

        :param change: Change.
        :raises RuntimeError: Reaction triggered during history redo/undo.
        """
        with self.app.write_context():

            # Check for inconsistent changes triggered during reactions while executing.
            if self.__executing:
                if isinstance(change, BaseAtomicChange) and change.history is not self:
                    error = "reaction triggered during history redo/undo in {}".format(
                        change.obj
                    )
                    raise RuntimeError(error)
                return

            # We should always be in a batch (the application should make sure of it).
            assert self._current_batches

            # Add to batch.
            with self._batch_context("Push Change", change=change):
                self.flush_redo()
                self._current_batches[-1]._changes.append(change)

    def __exit_batch__(self, batch):
        # type: (Batch) -> None
        """
        Exit batch context.

        :param batch: Batch.
        """
        with self.app.write_context():
            if self.__executing:
                return
            with self._batch_context("Exit Batch", batch=batch):
                assert batch is self._current_batches[-1].change
                self._current_batches[-1]._closed = True
                self._current_batches.pop()



# -*- coding: utf-8 -*-

import pytest

from objetto.utils.reraise_context import ReraiseContext


def test_dummy_context():
    def func_a():
        func_b()

    def func_b():
        func_c()

    def func_c():
        raise ValueError("something is wrong")

    with pytest.raises(ValueError, match="oops; something is wrong"):
        with ReraiseContext(ValueError, "oops"):
            func_a()


if __name__ == "__main__":
    pytest.main()
